'''Utilities for patching in cassettes'''

from .stubs import VCRHTTPConnection, VCRHTTPSConnection
from six.moves import http_client as httplib


# Save some of the original types for the purposes of unpatching
_HTTPConnection = httplib.HTTPConnection
_HTTPSConnection = httplib.HTTPSConnection

try:
    # Try to save the original types for requests
    import requests.packages.urllib3.connectionpool as cpool
    _VerifiedHTTPSConnection = cpool.VerifiedHTTPSConnection
    _cpoolHTTPConnection = cpool.HTTPConnection
    _cpoolHTTPSConnection = cpool.HTTPSConnection
except ImportError:  # pragma: no cover
    pass

try:
    # Try to save the original types for urllib3
    import urllib3
    _VerifiedHTTPSConnection = urllib3.connectionpool.VerifiedHTTPSConnection
except ImportError:  # pragma: no cover
    pass

try:
    # Try to save the original types for httplib2
    import httplib2
    _HTTPConnectionWithTimeout = httplib2.HTTPConnectionWithTimeout
    _HTTPSConnectionWithTimeout = httplib2.HTTPSConnectionWithTimeout
    _SCHEME_TO_CONNECTION = httplib2.SCHEME_TO_CONNECTION
except ImportError:  # pragma: no cover
    pass

try:
    # Try to save the original types for boto
    import boto.https_connection
    _CertValidatingHTTPSConnection = \
        boto.https_connection.CertValidatingHTTPSConnection
except ImportError:  # pragma: no cover
    pass


def install(cassette):
    """
    Patch all the HTTPConnections references we can find!
    This replaces the actual HTTPConnection with a VCRHTTPConnection
    object which knows how to save to / read from cassettes
    """
    httplib.HTTPConnection = VCRHTTPConnection
    httplib.HTTPSConnection = VCRHTTPSConnection
    httplib.HTTPConnection.cassette = cassette
    httplib.HTTPSConnection.cassette = cassette

    # patch requests v1.x
    try:
        import requests.packages.urllib3.connectionpool as cpool
        from .stubs.requests_stubs import VCRVerifiedHTTPSConnection
        cpool.VerifiedHTTPSConnection = VCRVerifiedHTTPSConnection
        cpool.VerifiedHTTPSConnection.cassette = cassette
        cpool.HTTPConnection = VCRHTTPConnection
        cpool.HTTPConnection.cassette = cassette
    # patch requests v2.x
        cpool.HTTPConnectionPool.ConnectionCls = VCRHTTPConnection
        cpool.HTTPConnectionPool.cassette = cassette
        cpool.HTTPSConnectionPool.ConnectionCls = VCRHTTPSConnection
        cpool.HTTPSConnectionPool.cassette = cassette
    except ImportError:  # pragma: no cover
        pass

    # patch urllib3
    try:
        import urllib3.connectionpool as cpool
        from .stubs.urllib3_stubs import VCRVerifiedHTTPSConnection
        cpool.VerifiedHTTPSConnection = VCRVerifiedHTTPSConnection
        cpool.VerifiedHTTPSConnection.cassette = cassette
        cpool.HTTPConnection = VCRHTTPConnection
        cpool.HTTPConnection.cassette = cassette
    except ImportError:  # pragma: no cover
        pass

    # patch httplib2
    try:
        import httplib2 as cpool
        from .stubs.httplib2_stubs import VCRHTTPConnectionWithTimeout
        from .stubs.httplib2_stubs import VCRHTTPSConnectionWithTimeout
        cpool.HTTPConnectionWithTimeout = VCRHTTPConnectionWithTimeout
        cpool.HTTPSConnectionWithTimeout = VCRHTTPSConnectionWithTimeout
        cpool.SCHEME_TO_CONNECTION = {
            'http': VCRHTTPConnectionWithTimeout,
            'https': VCRHTTPSConnectionWithTimeout
        }
    except ImportError:  # pragma: no cover
        pass

    # patch boto
    try:
        import boto.https_connection as cpool
        from .stubs.boto_stubs import VCRCertValidatingHTTPSConnection
        cpool.CertValidatingHTTPSConnection = VCRCertValidatingHTTPSConnection
        cpool.CertValidatingHTTPSConnection.cassette = cassette
    except ImportError:  # pragma: no cover
        pass


def reset():
    '''Undo all the patching'''
    httplib.HTTPConnection = _HTTPConnection
    httplib.HTTPSConnection = _HTTPSConnection
    try:
        import requests.packages.urllib3.connectionpool as cpool
        # unpatch requests v1.x
        cpool.VerifiedHTTPSConnection = _VerifiedHTTPSConnection
        cpool.HTTPConnection = _cpoolHTTPConnection
        # unpatch requests v2.x
        cpool.HTTPConnectionPool.ConnectionCls = _cpoolHTTPConnection
        cpool.HTTPSConnection = _cpoolHTTPSConnection
        cpool.HTTPSConnectionPool.ConnectionCls = _cpoolHTTPSConnection
    except ImportError:  # pragma: no cover
        pass

    try:
        import urllib3.connectionpool as cpool
        cpool.VerifiedHTTPSConnection = _VerifiedHTTPSConnection
        cpool.HTTPConnection = _HTTPConnection
        cpool.HTTPSConnection = _HTTPSConnection
        cpool.HTTPConnectionPool.ConnectionCls = _HTTPConnection
        cpool.HTTPSConnectionPool.ConnectionCls = _HTTPSConnection
    except ImportError:  # pragma: no cover
        pass

    try:
        import httplib2 as cpool
        cpool.HTTPConnectionWithTimeout = _HTTPConnectionWithTimeout
        cpool.HTTPSConnectionWithTimeout = _HTTPSConnectionWithTimeout
        cpool.SCHEME_TO_CONNECTION = _SCHEME_TO_CONNECTION
    except ImportError:  # pragma: no cover
        pass

    try:
        import boto.https_connection as cpool
        cpool.CertValidatingHTTPSConnection = _CertValidatingHTTPSConnection
    except ImportError:  # pragma: no cover
        pass



import six
from six import BytesIO
from six.moves.http_client import HTTPMessage
try:
    import http.client
except ImportError:
    pass


"""
The python3 http.client api moved some stuff around, so this is an abstraction
layer that tries to cope with this move.
"""


def get_header(message, name):
    if six.PY3:
        return message.getallmatchingheaders(name)
    else:
        return message.getheader(name)


def get_header_items(message):
    if six.PY3:
        return dict(message._headers).items()
    else:
        return message.dict.items()


def get_headers(response):
    for key in response.msg.keys():
        if six.PY3:
            yield key, response.msg.get_all(key)
        else:
            yield key, response.msg.getheaders(key)


def get_httpmessage(headers):
    if six.PY3:
        return http.client.parse_headers(BytesIO(headers))
    msg = HTTPMessage(BytesIO(headers))
    msg.fp.seek(0)
    msg.readheaders()
    return msg



from six.moves.urllib.parse import urlparse, urlencode, urlunparse
import copy


def _remove_headers(request, headers_to_remove):
    headers = copy.copy(request.headers)
    headers_to_remove = [h.lower() for h in headers_to_remove]
    keys = [k for k in headers if k.lower() in headers_to_remove]
    if keys:
        for k in keys:
            headers.pop(k)
        request.headers = headers
    return request


def _remove_query_parameters(request, query_parameters_to_remove):
    query = request.query
    new_query = [(k, v) for (k, v) in query
                 if k not in query_parameters_to_remove]
    if len(new_query) != len(query):
        uri_parts = list(urlparse(request.uri))
        uri_parts[4] = urlencode(new_query)
        request.uri = urlunparse(uri_parts)
    return request


def filter_request(
        request,
        filter_headers,
        filter_query_parameters,
        before_record,
        ignore_hosts
        ):
    request = copy.copy(request)  # don't mutate request object
    if hasattr(request, 'headers') and filter_headers:
        request = _remove_headers(request, filter_headers)
    if hasattr(request, 'host') and request.host in ignore_hosts:
        return None
    if filter_query_parameters:
        request = _remove_query_parameters(request, filter_query_parameters)
    if before_record:
        request = before_record(request)
    return request
